; SAT-TS (CNF)
; MPradella 20080304
; v. MzScheme 4
;
; --------------------------------------------------------------------------
;
; Copyright (C) 2006, 2007 Matteo Pradella (pradella@elet.polimi.it)
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
;
; --------------------------------------------------------------------------

#lang scheme/base

(provide
					; -- main --
 parse-it ; backward compatibility
 please-parse-it
 set-gamma!
 set-tiles!
 set-inverse-pi!
 toggle-dump-tex!
 toggle-dump-pict!
					; -- utilities ---
 tiles-of
 tiles-not-in
 big-picture-of
 union
 intersection
 impl
 only-one
 tiling-ec)

(require 
 scheme/system
 (only-in srfi/1 lset-union lset-intersection))


					; --- globals + interface ---

(define *aux-vars-counter* 0) ; auxiliary variables used for CNF
(define *tiles*           #f)
(define *gamma*           #f)
(define *inverse-pi*      #f)
(define *props*           #f)
(define *inv-props*       #f) ; inverse of *props* (for unparsing)
(define *dump-tex?*       #f)
(define *dump-pict?*      #f)

(define (set-tiles! tiles)
  (set! *tiles* tiles))
(define (set-gamma! gamma)
  (set! *gamma* gamma))
(define (set-inverse-pi! pi)
  (set! *inverse-pi* pi))

(define (toggle-dump-tex!)
  (set! *dump-tex?* (not *dump-tex?*)))

(define (toggle-dump-pict!)
  (set! *dump-pict?* (not *dump-pict?*)))

					; ------------
					; useful stuff
					; ------------

(define-syntax list-ref-ref  ; poor man's inlining...
  (syntax-rules ()
    ((list-ref-ref p i j)
     (if (null? p)
	 '?
	 (list-ref (list-ref p i) j)))))

(define (make-array-2d max-rows max-cols fill)
  (let ((vec (make-vector max-rows #f)))
    (let loop ((x 0))
      (if (< x max-rows)
	  (begin
	    (vector-set! vec x (make-vector max-cols fill))
	    (loop (+ x 1)))
	  vec))))

(define-syntax array-2d-set! 
  (syntax-rules ()
    ((_ the-array row col val)
     (vector-set! (vector-ref the-array (- row 1)) (- col 1) val))))

(define-syntax array-2d-ref 
  (syntax-rules ()
    ((_ the-array row col)
     (vector-ref (vector-ref the-array (- row 1)) (- col 1)))))


					; -----------------
					; Formula -> DIMACS
					; -----------------

(define-syntax simplify-proposition
  (syntax-rules ()
    ((simplify-proposition cl)
     (cond
      ((or (number? cl)(boolean? cl))
       cl)
      ((eq? 'not (car cl))
       (let ((val (cadr cl)))
	 (cond 
	  ((boolean? val)
	   (not val))
	  ((number? val)
	   (- val))
	  (else
	   (error "simplify-proposition" "bad arg not" val)))))
      (else 
       (error "simplify-proposition" "bad arg" cl))))))


(define-syntax simplify-clause
  (syntax-rules ()
    ((simplify-clause cl)
     (if (and (list? cl)(eq? 'or (car cl)))

	 (let loop ((c (cdr cl))
		    (clause '()))
	   (if (null? c)
	       clause
	       (let ((temp (simplify-proposition (car c))))
		 (if (not (eq? #t temp))
		     (loop (cdr c) (if (number? temp) 
				       (cons temp clause)
				       clause))
		     #f))))

					; case: clause == atom
	 (let ((val (simplify-proposition cl)))
	   (if (not (eq? val #t)) 
	       val
	       #f))))))


(define (toDIMACS f)
  "translates formula f to string into DIMACS format"

  (display "p cnf ")
  (display *aux-vars-counter*)
  (display " ")

  (let ((clause-number 0))
    (set! f 
	  (let loop ((c f)
		     (out '()))
	    (if (null? c)
		out
		(let ((temp (simplify-clause (car c))))
		  (when temp
			(set! clause-number (+ 1 clause-number)))
		  (loop (cdr c) (if temp (cons temp out) out))))))

    (display clause-number) ; # of clauses
    (newline)

    (for-each
     (lambda (c)
       (if (list? c)
	   (for-each 
	    (lambda (y) (display y)(display " "))
	    c)
	   (begin (display c)(display " ")))
       (display "0")(newline))
     f)))


					; ------------------------------
					; Data structures initialization
					; ------------------------------

(define (make-props m n picture)
  "define the propositions ((m,n) is the size of the pict)"
  (let ((counter 1))

    (set! *props*     (make-array-2d m n '())) 
    (set! *inv-props* (make-vector (* m n (length *gamma*)) "")) 

    (let loop ((i 1))
      (when (<= i m)
	    (let loop ((j 1))
	      (when (<= j n)
		    (if (or (= i 1)(= j 1)(= i m)(= j n))
			(begin
			  (array-2d-set! *props* i j (list (list '$ counter)))
			  (vector-set! *inv-props* counter '$)
			  (set! counter (+ 1 counter)))
			(for-each
			 (lambda (a)
			   (array-2d-set! *props* i j 
					  (cons (list a counter)
						(array-2d-ref *props* i j)))
			   (vector-set! *inv-props* counter a)
			   (set! counter (+ 1 counter)))
			 
			 (let ((pix (list-ref-ref picture (- i 1) (- j 1))))
			   (if (eq? pix '?)
			       *gamma*
			       (cdr (assoc pix *inverse-pi*))))))
		    (loop (+ j 1))))
	    (loop (+ i 1))))
    (vector-set! *inv-props* 0 counter) ; at position 0 we put the actual size
    (set! *aux-vars-counter* counter)))


					; ---------
					; Semantics
					; ---------

					; *Formula 1*
(define-syntax just-one
  (syntax-rules ()
    ((just-one items)
     (let* ((last-clause '())
	    (c1
	     (foldl
	      append '()
	      (map
	       (lambda (x)
		 (let ((sy      *aux-vars-counter*)
		       (not-it  (filter (lambda (z) (not (equal? x z))) items)))
		   
		   (set! *aux-vars-counter* (+ 1 *aux-vars-counter*))
		   (set! last-clause (cons sy last-clause))
		   
		   (append
		    (list
		     `(or ,sy (not ,x)  
			  ,@(map (lambda (y) y) 
				 not-it))
		     `(or (not ,sy) ,x))
		    
		    (map (lambda (x) 
			   `(or (not ,sy) (not ,x)))
			 not-it))))
	       items))))
       (append c1
	       (list (cons 'or last-clause)))))))


(define (pict-fma pict v-max h-max)
  "generates a propositional formula representing the picture pict"
  (let loop ((i 1)
	     (out '()))
    (if (> i v-max)
	out
	(begin
	  (let loop ((j 1))
	    (when (<= j h-max)
		  (let ((content (map (lambda (x)
					(cadr x))
				      (array-2d-ref *props* i j))))
		    (set! out (append out (just-one content)))
		    (loop (+ 1 j)))))
	  (loop (+ i 1) out)))))



					; *Formula 2*
(define-syntax just-tiles
  (syntax-rules ()
    ((just-tiles items)
     (let* ((last-clause '())
	    (c1 
	     (append
	      (foldl 
	       append '()
	       (map
		(lambda (x)
		  (let ((sy *aux-vars-counter*))
		    
		    (set! *aux-vars-counter* (+ 1 *aux-vars-counter*))
		    (set! last-clause (cons sy last-clause))
		    
		    (cons
		     `(or ,sy 
			  ,@(map (lambda (y) `(not ,y)) x))
		     
		     (map (lambda (z) 
			    `(or (not ,sy) ,z))
			  x))))
		items)))))
       (append c1
	       (list (cons 'or last-clause)))))))



(define (tiling-big-and m n)
					;generates a propositional formula representing the tiling constraints
  (define (the-tau i j the-tile)
    (let loop ((h 0)
	       (out '()))
      (if (> h 1)
	  out
	  (begin
	    (let loop ((k 0))
	      (when (< k 2)
		    (let ((content (assoc (list-ref-ref the-tile h k) 
					  (array-2d-ref *props* (+ i h) (+ j k)))))
		      (set! out (cons (if content
					  (cadr content)
					  #f)
				      out)))
		    (loop (+ 1 k))))
	    (loop (+ 1 h) out)))))

  (let loop ((i 1)
	     (out '()))
    (if (>= i m)
	out
	(begin
	  (let loop ((j 1))
	    (when (< j n)
		  (set! out (append (just-tiles
				     (map (lambda (t) (the-tau i j t))
					  *tiles*))
				    out))
		  (loop (+ j 1))))
	  (loop (+ i 1) out)))))


					; ------
					; Parser
					; ------

(define (the-parser filename . args)
					;args may be a list of lists representing a picture,
					;or a picture size (rows, columns) -- not counting borders
  (define v-max #f)
  (define h-max #f)
  (define fma1  #f)
  (define props #f)
  (define i-props #f)

  (display "This is SAT-TS v.20080304")
  (newline)
  (display "generating formula I...")

  (set! fma1 
	(case (length args)
	  ((2) 
	   (set! v-max (+ 2 (car args)))
	   (set! h-max (+ 2 (cadr args)))
	   (make-props v-max h-max '())
	   (pict-fma '() v-max h-max))
	  ((1)
	   (set! v-max (length (car args)))
	   (set! h-max (length (car (car args))))
	   (make-props v-max h-max (car args))
	   (pict-fma (car args) v-max h-max))
	  (else
	   (error "Args: either a picture or two integers"))))

  (newline)
  (display "generating formula II...")

  (let ((fma2 (tiling-big-and v-max h-max)))
    (newline)
    (display "dumping the output...")
    (newline)

    (with-output-to-file 
	filename 
      (lambda ()
	(toDIMACS (append fma1 fma2))) #:exists 'replace))) 


					; -------------
					; *Back-parser*
					; -------------

(define (unparse-it infile h-max)
  (define nv (vector-ref *inv-props* 0))

  (display "back-parsing...")

  (with-input-from-file 
      infile
    (lambda ()
      (let ((the-content
	     (if (or (eq? (read) 'SAT) 	        ; MiniSat
		     (eq? (read) 'SATISFIABLE)) ; Sato (+ discards prev. read)
		 (let loop ((x (read))(content '()))
		   (if (and (number? x)
			    (> (abs x) nv))
		       content
		       (loop
			(read)
			(if (and (number? x)(positive? x))
			    (append content (list x))
			    content))))
		 '())))

	(sat-to-pict the-content h-max)

	(when *dump-pict?* 
	      (display "dumping the picture...")
	      (newline)
	      (with-output-to-file 
		  "out.pict.txt"
		(lambda ()
		  (display (sat-dump-pict the-content h-max))) #:exists 'replace))

	(when *dump-tex?* 
	      (display "writing .tex table...")
	      (newline)
	      (with-output-to-file 
		  "out.tex"
		(lambda ()
		  (display (sat-to-tex the-content h-max))) #:exists 'replace))))))


(define (sat-to-pict str h-max)
  (define h (map (lambda (x)
		   (let* ((res (vector-ref *inv-props* x)))
		     ((cond
		       ((eq? res '$)  (lambda (x) ""))
		       ((string? res) (lambda (x) x))
		       ((number? res) number->string)
		       (else symbol->string)) res)))
		 str))
  (do ((cnt 1 (+ 1 cnt))
       (lst h (cdr lst)))
      ((null? lst) "")
    (display (car lst))
    (when (= 0 (remainder cnt h-max))
	  (display "\n"))))


(define (sat-dump-pict str h-max)
  "writes the parsed picture in Scheme notation on file"
  (display "((")
  (let ((h (map (lambda (x)
		  (let* ((res (vector-ref *inv-props* x)))
		    ((cond
		      ((eq? res '$)  (lambda (x) "$"))
		      ((string? res) (lambda (x) x))
		      ((number? res) number->string)
		      (else symbol->string)) res)))
		str)))
    (do ((cnt 1 (+ 1 cnt))
	 (lst h (cdr lst)))
	((null? lst) "")
      (display (car lst))
      (display " ")
      (when (= 0 (remainder cnt h-max))
	    (display ")\n")
	    (when (> (length lst) 2)
		  (display "("))))
    (display ")\n")))


(define (sat-to-tex str h-max)
  "writes the parsed picture in LaTeX notation on file"
  (let ((h (map (lambda (x)
		  (let* ((res (vector-ref *inv-props* x)))
		    ((cond
		      ((eq? res '$)  (lambda (x) "\\#"))
		      ((string? res) (lambda (x) x))
		      ((number? res) number->string)
		      (else symbol->string)) res)))
		str)))
    (do ((str "" (string-append str
				(car lst)
				(if (= 0 (remainder cnt h-max))
				    "\\\\ \n"
				    " & ")))
	 (cs  "{|c" (if flag
			cs
			(if (= 0 (remainder cnt h-max))
			    (begin
			      (set! flag #t)
			      (string-append cs "|}"))
			    (string-append cs "c"))))
	 (flag #f flag)
	 (cnt 1 (+ 1 cnt))
	 (lst h (cdr lst)))
	((null? lst)
	 (display "\\begin{array}")
	 (display cs) 
	 (display "\n\\hline\n")
	 (display str)
	 (display "\\hline\n\\end{array}\n")
	 ""))))


					; ------
					; *Main*
					; ------

(define (parse-it tiles gamma pi . args) ; for compatibility
  (set-tiles! tiles)
  (set-gamma! gamma)
  (set-inverse-pi! pi)
  (apply please-parse-it args))

(define (please-parse-it . args)
  (let ((outfile "out.txt"))
    (case (length args)
      ((1)
       (time (the-parser outfile (car args))))
      ((2)
       (time (the-parser outfile
			 (car args) (cadr args))))))

  (display "calling MiniSat...")(newline)  
  (collect-garbage)

  (if (not (system "minisat out.txt out.sat.txt"))
      (let ((h-max (vector-length (vector-ref *props* 0))))
	(unparse-it "out.sat.txt" h-max))
      (error "parse-it" "minisat problems?")))




					; --- Utilities for writing Tiling Systems ---


(define (tiles-of pict)
  "utility: extracts every 2x2 tiles from picture pict"
  (let ((m (length pict))
	(n (length (car pict))))

    (let loop ((i 0)
	       (out '()))
      (if (>= i (- m 1))
	  out
	  (begin
	    (let loop ((j 0))
	      (when (< j (- n 1))
		    (set! out (cons 
			       (list
				(list (list-ref (list-ref pict i) j)
				      (list-ref (list-ref pict i) (+ 1 j)))
				(list (list-ref (list-ref pict (+ 1 i)) j)
				      (list-ref (list-ref pict (+ 1 i)) (+ 1 j))))
			       out))
		    (loop (+ j 1))))
	    (loop (+ i 1) out))))))

(define (tiles-not-in set gamma)
  "utility: complement of the tiles in set"
  (let ((gamma1 (cons '$ gamma)))
    (let loop ((i gamma1)
	       (out '()))
      (if (null? i)
	  out
	  (begin
	    (let loop ((j gamma1))
	      (when (not (null? j))
		    (let loop ((k gamma1))
		      (when (not (null? k))
			    (let loop ((l gamma1))
			      (when (not (null? l))
				    (let ((ii (car i))
					  (jj (car j))
					  (kk (car k))
					  (ll (car l)))
				      (unless (or
					       (member `((,ii ,jj)(,kk ,ll)) set)
					       (and (eq? ii '$)(eq? jj '$)(eq? kk '$)(eq? ll '$))
					       (and (eq? ii '$)(eq? ll '$)(not (eq? jj '$))(not (eq? kk '$)))
					       (and (eq? jj '$)(eq? kk '$)(not (eq? ii '$))(not (eq? ll '$))))
					      (set! out (cons `((,ii ,jj)(,kk ,ll)) out))))
				    (loop (cdr l))))
			    (loop (cdr k))))
		    (loop (cdr j))))
	    (loop (cdr i) out))))))


(define (big-picture-of side char)

  (let ((top (list 
	      (let loop ((i 0)(out '()))
		(if (= i (+ 2 side))
		    out
		    (loop (+ i 1) (cons '$ out)))))))
    (append
     top
     (let loop ((j 0)(out '()))
       (if (= j side)
	   out
	   (loop (+ j 1) (cons (append
				'($)
				(let loop ((k 0)(out '()))
				  (if (= k side) 
				      out
				      (loop (+ k 1) (cons char out))))
				'($))
			       out))))
     top)))



(define (union . sets)
  (apply lset-union (cons equal? sets))) 

(define (intersection . sets)
  (apply lset-intersection (cons equal? sets)))

(define (impl a b)
  (or (not a) b))

(define (only-one . elements)
  (do ((l elements
	  (cdr l))
       (sum 0
	    (if (car l)
		(+ sum 1)
		sum)))
      ((null? l)
       (= sum 1))))

(define (tiling-ec the-cond the-gamma) ; the-cond must be a lambda(i j k l)
  (let ((out '())
	(gamma (cons '$ the-gamma)))
    (for-each
     (lambda (i)
       (for-each
	(lambda(j)
	  (for-each
	   (lambda (k)
	     (for-each
	      (lambda (l)
		(when (the-cond i j k l)
		      (set! out 
			    (append out 
				    (list (list (list i j)(list k l)))))))
	      gamma))
	   gamma))
	gamma))
     gamma)
    out))

